@@ -10,7 +10,16 @@
  * a minor inefficiency. for extensibility, the implementation uses 
  * a separate function for each distance measure.
  * 
- * ceeboo 2005
+ * fixed 'bug' due to the copy on write mechanism that was sneaked
+ * in by R 2.3.0 or 2.3.1. this is a tell tale reminder that we
+ * should not rely on anything in R :-( this, of course, makes
+ * things less structured, maintainable and, efficient.
+ *
+ * fixme: 1) conceptually we should return proper dist objects here
+ *        instead of piecing it together on the R level. however,
+ *        see remark above. 2) complex data are not supported.
+ * 
+ * ceeboo 2005, 2006
  */
 
 /* calculate Minkowsky distances (including one-norm) between the 
@@ -22,7 +31,7 @@
 SEXP pdist(SEXP R_x, SEXP R_y, SEXP R_p) {
 
     int nc, nx, ny;
-    int i, j, k, l, m;
+    int i, j, k, l, m, as_matrix = 0;
 
     double p, d, q, z;
     double *x, *y;
@@ -33,6 +42,8 @@ SEXP pdist(SEXP R_x, SEXP R_y, SEXP R_p) {
 
     if (R_y == R_NilValue)
        R_y = R_x;
+    else
+       as_matrix = 1;
 		    
     if (INTEGER(GET_DIM(R_y))[1] != nc)
        error("pdist: invalid number of columns");
@@ -51,17 +62,17 @@ SEXP pdist(SEXP R_x, SEXP R_y, SEXP R_p) {
     x = REAL(R_x);
     y = REAL(R_y);
 	
-    if (R_x == R_y)
+    if (!as_matrix && R_x == R_y)
        PROTECT(R_obj = NEW_NUMERIC(nx*(nx-1)/2));
     else
-       PROTECT(R_obj = NEW_NUMERIC(nx*ny));
+       PROTECT(R_obj = allocMatrix(REALSXP, nx, ny));
 		    
   
     q = 1 / p;
  
     m = 0;
     for (j = 0; j < ny; j++) {
-	if (R_x == R_y)
+	if (!as_matrix && R_x == R_y)
 	   i = j + 1;
         else
 	   i = 0;
@@ -82,19 +93,7 @@ SEXP pdist(SEXP R_x, SEXP R_y, SEXP R_p) {
 	    else
 	       REAL(R_obj)[m++] = NA_REAL;
 	}
-    }
-    
-    if (R_x != R_y) {
-	    
-       SEXP R_tmp;
-
-       PROTECT(R_tmp = NEW_INTEGER(2));
-
-       INTEGER(R_tmp)[0] = nx;
-       INTEGER(R_tmp)[1] = ny;
-	
-       SET_DIM(R_obj, R_tmp);
-       UNPROTECT(1);
+	R_CheckUserInterrupt();
     }
     
     UNPROTECT(1);
@@ -108,7 +107,7 @@ SEXP pdist(SEXP R_x, SEXP R_y, SEXP R_p) {
 SEXP mdist(SEXP R_x, SEXP R_y) {
 
     int nc, nx, ny; 
-    int i, j, k, l, m;
+    int i, j, k, l, m, as_matrix = 0;
     
     double d, z;
     double *x, *y;
@@ -117,8 +116,10 @@ SEXP mdist(SEXP R_x, SEXP R_y) {
 
     nc = INTEGER(GET_DIM(R_x))[1];
 
-    if (R_y == R_NilValue)
+    if (R_y == R_NilValue) 
        R_y = R_x;
+    else
+       as_matrix = 1;
 		    
     if (INTEGER(GET_DIM(R_y))[1] != nc)
        error("mdist: invalid number of columns");
@@ -129,14 +130,14 @@ SEXP mdist(SEXP R_x, SEXP R_y) {
     x = REAL(R_x);
     y = REAL(R_y);
 	
-    if (R_x == R_y)
+    if (!as_matrix && R_x == R_y)
        PROTECT(R_obj = NEW_NUMERIC(nx*(nx-1)/2));
     else
-       PROTECT(R_obj = NEW_NUMERIC(nx*ny));
+       PROTECT(R_obj = allocMatrix(REALSXP, nx, ny));
 		    
     m = 0;
     for (j = 0; j < ny; j++) {
-	if (R_x == R_y)
+	if (!as_matrix && R_x == R_y)
 	   i = j + 1;
 	else
 	   i = 0;
@@ -158,21 +159,9 @@ SEXP mdist(SEXP R_x, SEXP R_y) {
 	    else
 	       REAL(R_obj)[m++] = NA_REAL;
 	}
+	R_CheckUserInterrupt();
     }
        
-    if (R_x != R_y) {
-	    
-       SEXP R_tmp;
-
-       PROTECT(R_tmp = NEW_INTEGER(2));
-
-       INTEGER(R_tmp)[0] = nx;
-       INTEGER(R_tmp)[1] = ny;
-	
-       SET_DIM(R_obj, R_tmp);
-       UNPROTECT(1);
-    }
-    
     UNPROTECT(1);
 
     return R_obj;			    
@@ -184,7 +173,7 @@ SEXP mdist(SEXP R_x, SEXP R_y) {
 SEXP cdist(SEXP R_x, SEXP R_y) {
     
     int nc, nx, ny; 
-    int i, j, k, l, m;
+    int i, j, k, l, m, as_matrix = 0;
     
     double d, z;
     double *x, *y;
@@ -195,6 +184,8 @@ SEXP cdist(SEXP R_x, SEXP R_y) {
 
     if (R_y == R_NilValue)
        R_y = R_x;
+    else
+       as_matrix = 1;
 		    
     if (INTEGER(GET_DIM(R_y))[1] != nc)
        error("cdist: invalid number of columns");
@@ -205,14 +196,14 @@ SEXP cdist(SEXP R_x, SEXP R_y) {
     x = REAL(R_x);
     y = REAL(R_y);
 	
-    if (R_x == R_y)
+    if (!as_matrix && R_x == R_y)
        PROTECT(R_obj = NEW_NUMERIC(nx*(nx-1)/2));
     else
-       PROTECT(R_obj = NEW_NUMERIC(nx*ny));
+       PROTECT(R_obj = allocMatrix(REALSXP, nx, ny));
 		    
     m = 0; 
     for (j = 0; j < ny; j++) {
-	if (R_x == R_y)
+	if (!as_matrix && R_x == R_y)
 	   i = j + 1;
 	else
 	   i = 0;
@@ -233,21 +224,9 @@ SEXP cdist(SEXP R_x, SEXP R_y) {
 	    else
 	       REAL(R_obj)[m++] = NA_REAL;
 	}
+	R_CheckUserInterrupt();
     }
 
-    if (R_x != R_y) {
-
-       SEXP R_tmp;
-
-       PROTECT(R_tmp = NEW_INTEGER(2));
-
-       INTEGER(R_tmp)[0] = nx;
-       INTEGER(R_tmp)[1] = ny;
-	
-       SET_DIM(R_obj, R_tmp);
-       UNPROTECT(1);
-    }
-    
     UNPROTECT(1);
 
     return R_obj;
@@ -279,7 +258,7 @@ SEXP cdist(SEXP R_x, SEXP R_y) {
 SEXP bdist(SEXP R_x, SEXP R_y) {
 
     int nc, nx, ny; 
-    int i, j, k, l, t;
+    int i, j, k, l, t, as_matrix = 0;
     
     int *x, *y, *s;
 
@@ -291,6 +270,8 @@ SEXP bdist(SEXP R_x, SEXP R_y) {
     
     if (R_y == R_NilValue)
        R_y = R_x;
+    else
+       as_matrix = 1;
     
     if (INTEGER(GET_DIM(R_y))[1] != nc)
        error("bdist: invalid number of columns");
@@ -303,10 +284,10 @@ SEXP bdist(SEXP R_x, SEXP R_y) {
     
     s = Calloc(nx, int);
     
-    if (R_x != R_y)
-       PROTECT(R_obj = NEW_NUMERIC(nx*ny));
-    else
+    if (!as_matrix && R_x == R_y)
        PROTECT(R_obj = NEW_NUMERIC(nx*(nx-1)/2));
+    else
+       PROTECT(R_obj = allocMatrix(REALSXP, nx, ny));
    
     for (i = 0; i < nx; i++) {
 	t = 0;
@@ -319,7 +300,7 @@ SEXP bdist(SEXP R_x, SEXP R_y) {
     }
     l = 0;
     for (j = 0; j < ny; j++) {
-	if (R_x != R_y) {
+	if (as_matrix || R_x != R_y) {
 	   t = 0;
 	   for (k = 0; k < nc; k++) {
 	       if (y[j+k*ny] == NA_LOGICAL)
@@ -345,23 +326,11 @@ SEXP bdist(SEXP R_x, SEXP R_y) {
 	    else
 	       REAL(R_obj)[l++] = z;
 	}
+	R_CheckUserInterrupt();
     }
    
     Free(s);
    
-    if (R_x != R_y) {
-	    
-       SEXP R_tmp;
-
-       PROTECT(R_tmp = NEW_INTEGER(2));
-
-       INTEGER(R_tmp)[0] = nx;
-       INTEGER(R_tmp)[1] = ny;
-
-       SET_DIM(R_obj, R_tmp);
-       UNPROTECT(1);
-    }
-    
     UNPROTECT(1);
 
     return R_obj;
@@ -378,7 +347,7 @@ SEXP bdist(SEXP R_x, SEXP R_y) {
 SEXP ebdist(SEXP R_x, SEXP R_y) {
     
     int nc, nx, ny; 
-    int i, j, k, l, m;
+    int i, j, k, l, m, as_matrix = 0;
     
     double t, z;
     double *x, *y, *s;
@@ -389,6 +358,8 @@ SEXP ebdist(SEXP R_x, SEXP R_y) {
 
     if (R_y == R_NilValue)
        R_y = R_x;
+    else
+       as_matrix = 1;
 		    
     if (INTEGER(GET_DIM(R_y))[1] != nc)
        error("ebdist: invalid number of columns");
@@ -401,10 +372,10 @@ SEXP ebdist(SEXP R_x, SEXP R_y) {
 	
     s = Calloc(nx, double);
     
-    if (R_x == R_y)
+    if (!as_matrix && R_x == R_y)
        PROTECT(R_obj = NEW_NUMERIC(nx*(nx-1)/2));
     else
-       PROTECT(R_obj = NEW_NUMERIC(nx*ny));
+       PROTECT(R_obj = allocMatrix(REALSXP, nx, ny));
 		    
     for (i = 0; i < nx; i++) {
 	z = 0;
@@ -422,7 +393,7 @@ SEXP ebdist(SEXP R_x, SEXP R_y) {
     }
     m = 0; 
     for (j = 0; j < ny; j++) {
-	if (R_x == R_y) {
+	if (!as_matrix && R_x == R_y) {
 	   t = s[j];
 	   i = j + 1;
 	}
@@ -464,23 +435,11 @@ SEXP ebdist(SEXP R_x, SEXP R_y) {
 	    else
 	       REAL(R_obj)[m++] = NA_REAL;
 	}
+	R_CheckUserInterrupt();
     }
 
     Free(s);
     
-    if (R_x != R_y) {
-
-       SEXP R_tmp;
-
-       PROTECT(R_tmp = NEW_INTEGER(2));
-
-       INTEGER(R_tmp)[0] = nx;
-       INTEGER(R_tmp)[1] = ny;
-	
-       SET_DIM(R_obj, R_tmp);
-       UNPROTECT(1);
-    }
-    
     UNPROTECT(1);
 
     return R_obj;
@@ -497,7 +456,7 @@ SEXP ebdist(SEXP R_x, SEXP R_y) {
 SEXP fbdist(SEXP R_x, SEXP R_y) {
     
     int nc, nx, ny; 
-    int i, j, k, l, m;
+    int i, j, k, l, m, as_matrix = 0;
     
     double x1, x2, z1, z2;
     double *x, *y;
@@ -508,6 +467,8 @@ SEXP fbdist(SEXP R_x, SEXP R_y) {
 
     if (R_y == R_NilValue)
        R_y = R_x;
+    else
+       as_matrix = 1;
 		    
     if (INTEGER(GET_DIM(R_y))[1] != nc)
        error("fbdist: invalid number of columns");
@@ -518,14 +479,14 @@ SEXP fbdist(SEXP R_x, SEXP R_y) {
     x = REAL(R_x);
     y = REAL(R_y);
 	
-    if (R_x == R_y)
+    if (!as_matrix && R_x == R_y)
        PROTECT(R_obj = NEW_NUMERIC(nx*(nx-1)/2));
     else
-       PROTECT(R_obj = NEW_NUMERIC(nx*ny));
+       PROTECT(R_obj = allocMatrix(REALSXP, nx, ny));
 		    
     m = 0; 
     for (j = 0; j < ny; j++) {
-	if (R_x == R_y) 
+	if (!as_matrix && R_x == R_y) 
 	   i = j + 1;
 	else 
 	   i = 0;
@@ -557,21 +518,9 @@ SEXP fbdist(SEXP R_x, SEXP R_y) {
 	    else
 	       REAL(R_obj)[m++] = NA_REAL;
 	}
+	R_CheckUserInterrupt();
     }
 
-    if (R_x != R_y) {
-
-       SEXP R_tmp;
-
-       PROTECT(R_tmp = NEW_INTEGER(2));
-
-       INTEGER(R_tmp)[0] = nx;
-       INTEGER(R_tmp)[1] = ny;
-	
-       SET_DIM(R_obj, R_tmp);
-       UNPROTECT(1);
-    }
-    
     UNPROTECT(1);
 
     return R_obj;
@@ -581,7 +530,7 @@ SEXP fbdist(SEXP R_x, SEXP R_y) {
 SEXP adist(SEXP R_x, SEXP R_y) {
     
     int nc, nx, ny; 
-    int i, j, k, l, m;
+    int i, j, k, l, m, as_matrix = 0;
     
     double t, z;
     double *x, *y, *s;
@@ -592,6 +541,8 @@ SEXP adist(SEXP R_x, SEXP R_y) {
 
     if (R_y == R_NilValue)
        R_y = R_x;
+    else
+       as_matrix = 1;
 		    
     if (INTEGER(GET_DIM(R_y))[1] != nc)
        error("adist: invalid number of columns");
@@ -604,10 +555,10 @@ SEXP adist(SEXP R_x, SEXP R_y) {
 	
     s = Calloc(nx, double);
     
-    if (R_x == R_y)
+    if (!as_matrix && R_x == R_y)
        PROTECT(R_obj = NEW_NUMERIC(nx*(nx-1)/2));
     else
-       PROTECT(R_obj = NEW_NUMERIC(nx*ny));
+       PROTECT(R_obj = allocMatrix(REALSXP, nx, ny));
 		    
     #define EPS 1E-16
 
@@ -627,7 +578,7 @@ SEXP adist(SEXP R_x, SEXP R_y) {
     }
     m = 0; 
     for (j = 0; j < ny; j++) {
-	if (R_x == R_y) {
+	if (!as_matrix && R_x == R_y) {
 	   t = s[j];
 	   i = j + 1;
 	}
@@ -673,23 +624,11 @@ SEXP adist(SEXP R_x, SEXP R_y) {
 	    else
 	       REAL(R_obj)[m++] = NA_REAL;
 	}
+	R_CheckUserInterrupt();
     }
 
     Free(s);
     
-    if (R_x != R_y) {
-
-       SEXP R_tmp;
-
-       PROTECT(R_tmp = NEW_INTEGER(2));
-
-       INTEGER(R_tmp)[0] = nx;
-       INTEGER(R_tmp)[1] = ny;
-	
-       SET_DIM(R_obj, R_tmp);
-       UNPROTECT(1);
-    }
-    
     UNPROTECT(1);
 
     return R_obj;
@@ -699,11 +638,15 @@ SEXP adist(SEXP R_x, SEXP R_y) {
  * we allow only one subset index. for the subsripting
  * to work we need a proper array (see last argument).
  * 
- * ceeboo 2005
+ * note that subscriptArray returns indexes that start
+ * at 1 instead of zero. Yuck!
+ * 
+ * ceeboo 2005, 2006
  */
 
 SEXP subset_dist(SEXP R_x, SEXP R_s, SEXP R_l) {
-
+    if (TYPEOF(R_x) != REALSXP)
+	error("invalid data parameter");
     int n, m;
     int i, ii, j, jj, k;
 
@@ -715,13 +658,23 @@ SEXP subset_dist(SEXP R_x, SEXP R_s, SEXP R_l) {
     
     PROTECT(R_s = arraySubscript(0, R_s, GET_DIM(R_l), getAttrib, 
 						       (STRING_ELT), R_l));
-	
+    if (TYPEOF(R_s) != INTSXP)
+	error("arraySubscript returned unsupported data type");
+
     n = 1 + (int) sqrt(2*LENGTH(R_x));  /* number of elements */
     
+    if (LENGTH(R_x) != n*(n-1)/2)
+	error("invalid length");
+    
     m = LENGTH(R_s);			/* size of subset */
     
+    if (m < 2)
+	error("invalid subscripts");
+    for (int k = 0; k < m; k++)
+	if (INTEGER(R_s)[k] == NA_INTEGER)
+	    error("invalid subscripts");
+		
     s = INTEGER(R_s);
-
     x = REAL(R_x);
 
     PROTECT(R_obj = NEW_NUMERIC(m*(m-1)/2));
@@ -744,26 +697,261 @@ SEXP subset_dist(SEXP R_x, SEXP R_s, SEXP R_l) {
 	}
     }
     
-    /* setting the labels here is more efficient */
-
-    R_l = GET_DIMNAMES(R_l);
-    
-    if (R_l != R_NilValue) {
+    /* it is more secure to do this here */
 
+    if (!isNull((R_l = GetRowNames(GET_DIMNAMES(R_l))))) {
        SEXP R_str;
        
        PROTECT(R_str = NEW_STRING(m));
-       R_l = VECTOR_ELT(R_l, 0);
        for (k = 0; k < m; k++)
-	   SET_ELEMENT(R_str, k, mkChar(CHAR(VECTOR_ELT(R_l, s[k]-1))));
+	   SET_STRING_ELT(R_str, k, duplicate(STRING_ELT(R_l, s[k]-1)));
 
        setAttrib(R_obj, install("Labels"), R_str);
        UNPROTECT(1);
     }
-    
+    setAttrib(R_obj, install("Size"), ScalarInteger(m));
+
     UNPROTECT(2);
 
     return R_obj;
 }
 
+/* compute the rowSums for an R dist object. due to
+ * symmetry this is equivalent to colSums. rowMeans
+ * are not implemented as this can be trivially
+ * obtained from the values of rowSums.
+ * 
+ * na_rm implements the usual meaning of omitting 
+ * NA and NaN values.
+ *
+ * ceeboo 2006
+ */
+
+SEXP rowSums_dist(SEXP x, SEXP na_rm) {
+    if (TYPEOF(x) != REALSXP)
+        error("invalid data parameter");
+    if (TYPEOF(na_rm) != LGLSXP)
+        error("invalid option paramter");
+    int k, n;
+    SEXP r;
+
+    n = 1 + (int) sqrt(2*LENGTH(x));
+    
+    if (LENGTH(x) != n*(n-1)/2)
+        error("invalid length");
+    
+    PROTECT(r = allocVector(REALSXP, n));
+    
+    for (int i = 0; i < n; i++)
+	REAL(r)[i] = 0;
+    
+    k = 0;
+    for (int i = 0; i < n-1; i++) 
+        for (int j = i+1; j < n; j++) {
+            double z = REAL(x)[k++];
+            if (!R_FINITE(z)) {
+               if (ISNAN(z)) {
+                  if (LOGICAL(na_rm)[0] == TRUE)
+                     continue;
+                  if (ISNA(z))
+                     REAL(r)[i] = REAL(r)[j] = NA_REAL;
+                  else
+                     REAL(r)[i] = REAL(r)[j] = R_NaN;
+                } else
+                  REAL(r)[i] = REAL(r)[j] = z;
+               break;
+            }
+            REAL(r)[i] += z;
+	    REAL(r)[j] += z;
+        }
+    
+    UNPROTECT(1);
+    return r;
+}
+
+/* compute auto- or cross-distances with a user-supplied
+ * function given matrix data. this is experimental code.
+ *
+ * fixme: eval environment?
+ * 
+ * ceeboo 2006
+ */
+
+SEXP apply_dist(SEXP p) {
+    int k, n, nx, ny, as_matrix = 0;
+    SEXP r, c, tx, ty;
+    SEXP x, y, f;
+
+    p = CDR(p);
+    if (length(p) < 3)
+	error("invalid number of arguments");
+    x = CAR(p); y = CADR(p);
+    if (!isMatrix(x) || (!isNull(y) && !isMatrix(y)))
+	error("invalid data parameter(s)");
+    p = CDDR(p); f = CAR(p); 
+    if (!isFunction(f))
+	error("invalid function parameter");
+    p = CDR(p);
+
+    if (isNull(y))
+	y = x;	
+    else
+       as_matrix = 1;
+    
+    if ((n = INTEGER(GET_DIM(x))[1]) != INTEGER(GET_DIM(y))[1])
+	error("data parameters do not conform");
+    
+    nx = INTEGER(GET_DIM(x))[0];
+    ny = INTEGER(GET_DIM(y))[0];
+
+    if (!as_matrix && x == y)
+	PROTECT(r = allocVector(REALSXP, nx*(nx-1)/2));
+    else
+	PROTECT(r = allocMatrix(REALSXP, nx, ny));
+    
+    PROTECT(tx = allocVector(REALSXP, n)); 
+    PROTECT(ty = allocVector(REALSXP, n));
+
+    c = LCONS(f, LCONS(tx, LCONS(ty, p)));
+    
+    k = 0;
+    for (int j = 0; j < ny; j++) {
+	for (int k = 0; k < n; k++)
+	    REAL(ty)[k] = REAL(y)[j+k*ny];
+	for (int i = ((!as_matrix && x == y) ? j+1 : 0); i < nx; i++) {
+	    for (int k = 0; k < n; k++)
+		REAL(tx)[k] = REAL(x)[i+k*nx];
+	    SEXP s = eval(c, R_GlobalEnv);
+	    if (TYPEOF(s) != REALSXP || LENGTH(s) != 1)
+		error("invalid return value");
+	    REAL(r)[k++] = REAL(s)[0];
+	}			
+	R_CheckUserInterrupt();
+    }
+
+    UNPROTECT(3);
+    return r;
+}
+
+/* cluster_dist
+ *
+ * cluster an undirected graph as representable by an R dist object,
+ * i.e. find all the disconnected subgraphs of graph.
+ * 
+ * as input we expect R_x the vector storage representation of the 
+ * upper/lower triangle of a distance matrix (see dist) and R_beta
+ * the distance threshold.
+ *
+ * returns a factor of cluster labels (integers 1,2, ..., k, with
+ * k the number of clusters).
+ *
+ * NA or NaN distance values are interpreted as no link! this is a 
+ * simplification as we do not want to check for the 2^k possible 
+ * clusterings given each indeterminate link is actually either above 
+ * or below the threshold.
+ * 
+ * NA or NaN threshold values result in an error as we do not want 
+ * to to check for all the possible clusterings given the threshold 
+ * assumes a value in the range of the distances.
+ * 
+ * fixme: 1) can we do this in less than O(n^2) time?
+ *	  2) can we use a strict threshold? 
+ * 
+ * ceeboo 2006
+ */
+
+SEXP cluster_dist(SEXP R_x, SEXP R_beta) {
+    if (TYPEOF(R_x) != REALSXP)
+	error("cluster_dist: 'x' invalid storage type");
+    if (TYPEOF(R_beta) != REALSXP)
+	error("cluster_dist: 'beta' invalid storage type");
+    int i, j, k, l, n, o, na, *c, *b;
+    char *s;
+    double beta, *x;
+
+    SEXP R_str, R_obj;
+
+    n = (int) sqrt(2 * length(R_x)) + 1;
+
+    if (n < 3 || n * (n - 1) / 2 != length(R_x))
+       error("cluster_dist: 'x' invalid length");
+   
+    beta = REAL(R_beta)[0];	    /* distance threshold */
+
+    if (ISNAN(beta))
+       error("cluster_dist: 'beta' NA or NaN");
+	       
+    PROTECT(R_obj = NEW_INTEGER(n));
+    c = INTEGER(R_obj);
+
+    for (i = 0; i < n; i++)
+	c[i] = i;
+    
+    x = REAL(R_x);
+    
+    k = na = 0;
+    for (i = 0; i < n - 1; i++)
+	for (j = i + 1; j < n; j++) {
+	    if (ISNAN(x[k])) {
+	       na++;
+	       continue;
+	    }
+	    if (beta >= x[k++]) {
+	       if (c[j] == c[i])
+		  continue;
+	       if (c[j] == j)
+		  c[j] = c[i];
+	       else {
+		  o = c[j];
+		  for (l = 0; l < n; l++)
+		      if (c[l] == o)
+		         c[l] = c[i];
+	       }
+	    }
+	}
+    if (na)
+       warning("cluster_dist: found NA (NaN) distance values, different solutions may be possible.");
+    
+    /* make indexes contiguous */
+    
+    b = Calloc(n, int);
+    
+    k = 0;
+    for (i = 0; i < n; i++) {
+	j = c[i];
+	if (b[j] == 0)
+	   b[j] = ++k;
+	c[i] = b[j];
+    }
+    Free(b);
+    
+    /* make return value a factor */
+    
+    s = Calloc(k/10+2, char);           /* stringified integers */
+    
+    PROTECT(R_str = NEW_STRING(k));
+    for (j = 0; j < k; j++) {
+        sprintf(s,"%i",j+1);
+        SET_STRING_ELT(R_str, j, mkChar(s));
+    }
+    Free(s);
+ 
+    SET_LEVELS(R_obj, R_str);
+    UNPROTECT(1);
+
+    PROTECT(R_str = NEW_STRING(1));
+    SET_STRING_ELT(R_str, 0, mkChar("factor"));
+	                            
+    SET_CLASS(R_obj, R_str);
+    UNPROTECT(1);
+		
+    /* we are done */
+	    
+    UNPROTECT(1);
+    
+    return R_obj;
+}
+
+
+
 /* the end */
